
using Toybox.Application as App;
using Toybox.Background as Bg;
using Toybox.System as Sys;
using Toybox.WatchUi as Ui;
using Toybox.Time;

// In-memory current location.
// Previously persisted in App.Storage, but now persisted in Object Store due to #86 workaround for App.Storage firmware bug.
// Current location retrieved/saved in checkPendingWebRequests().
// Persistence allows weather and sunrise/sunset features to be used after watch face restart, even if watch no longer has current
// location available.
//var gLocationLat = null;
//var gLocationLng = null;


(:background)
class IkigaiApp extends App.AppBase {

	var mView;
	var mFieldTypes = new [3];

	function initialize() {

		AppBase.initialize();

	}

	/*
	// onStart() is called on application start up
	function onStart(state) {
	}

	// onStop() is called when your application is exiting
	function onStop(state) {
	}
	*/

	// Return the initial view of your application here
	function getInitialView() {
		System.println("Initialisation  true");
		setProperty("initialisation", true);
		mView = new CrystalView();
		onSettingsChanged(); // After creating view.
		
		setProperty("language" ,WatchUi.loadResource( Rez.Strings.language ));
		setProperty("hasFloors" ,WatchUi.loadResource( Rez.Strings.hasFloors ));
		setProperty("garminDevice" ,WatchUi.loadResource( Rez.Strings.garminDevice ));
		setProperty("hasUniqueIdentifier" ,WatchUi.loadResource( Rez.Strings.hasUniqueIdentifier ));
		    
		return [mView];
	}

	function getView() {
		return mView;
	}

	// New app settings have been received so trigger a UI update
	function onSettingsChanged() {
		mFieldTypes[0] = getProperty("Field1Type");
		mFieldTypes[1] = getProperty("Field2Type");
		mFieldTypes[2] = getProperty("Field3Type");

		mView.onSettingsChanged(); // Calls checkPendingWebRequests().

		Ui.requestUpdate();
	}

	function hasField(fieldType) {
		return ((mFieldTypes[0] == fieldType) ||
			(mFieldTypes[1] == fieldType) ||
			(mFieldTypes[2] == fieldType));
	}


	// Determine if any web requests are needed.
	// If so, set approrpiate pendingWebRequests flag for use by BackgroundService, then register for
	// temporal event.
	// Currently called on layout initialisation, when settings change, and on exiting sleep.
	(:background_method)
	function checkPendingWebRequests() {
		
		System.println("checkPendingWebRequest");
		if (!(Sys has :ServiceDelegate)) {
			return;
		}

		var pendingWebRequests = getProperty("PendingWebRequests");
		if (pendingWebRequests == null) {
			pendingWebRequests = {};
		}
		
		var lastRefreshIkigaiStatus = getProperty("lastRefreshIkigaiStatus");
		if (lastRefreshIkigaiStatus == null || (Time.now().value() > (lastRefreshIkigaiStatus + 300)))
		{
			lastRefreshIkigaiStatus = Time.now().value();
			setProperty("lastRefreshIkigaiStatus",lastRefreshIkigaiStatus);
			pendingWebRequests["ikigaiStatus"] = true; //Say that we want a refresh of IkigaiStatus
		} 
		
		
		System.println("pendingWebRequests.keys().size() =" + pendingWebRequests.keys().size() );
		// If there are any pending requests:
		if (pendingWebRequests.keys().size() > 0) {

			// Register for background temporal event as soon as possible.
			var lastTime = Bg.getLastTemporalEventTime();
			if (lastTime !=null)
			{
				if (getProperty("initialisation")==false) {
					System.println("last refresh = " + lastTime.toString());
					// Events scheduled for a time in the past trigger immediately.
					//var nextTime = lastTime.add(new Time.Duration(180 * 60));
					var nextTime = lastTime.add(new Time.Duration(5 * 60));

					Bg.registerForTemporalEvent(nextTime);
				} else{
					System.println("Initialisation for registerTemporalEvent");
					var nextTime = lastTime.add(new Time.Duration(5 * 60));
					Bg.registerForTemporalEvent(nextTime);

				}
			} else {
				Bg.registerForTemporalEvent(Time.now());
				App.getApp().setProperty("nbrFailedConnection",0);
				App.getApp().setProperty("nbrFailedConnectionWithoutInit",0);
			}
		}

		setProperty("PendingWebRequests", pendingWebRequests);
		

		   
	}

	(:background_method)
	function getServiceDelegate() {
		System.println("launch backgroundservice");
		return [new BackgroundService()];
	}

	// Handle data received from BackgroundService.
	// On success, clear appropriate pendingWebRequests flag.
	// data is Dictionary with single key that indicates the data type received. This corresponds with Object Store and
	// pendingWebRequests keys.
	(:background_method)
	function onBackgroundData(data) {
		var pendingWebRequests = getProperty("PendingWebRequests");
		if (pendingWebRequests == null) {
			//Sys.println("onBackgroundData() called with no pending web requests!");
			pendingWebRequests = {};
		}

		var type = data.keys()[0]; // Type of received data.
		var storedData = getProperty(type);
		var receivedData = data[type]; // The actual data received: strip away type key.
		
		// No value in showing any HTTP error to the user, so no need to modify stored data.
		// Leave pendingWebRequests flag set, and simply return early.
		System.println("Data received");


		// New data received: clear pendingWebRequests flag and overwrite stored data.
		storedData = receivedData;
		pendingWebRequests.remove(type);
		setProperty("PendingWebRequests", pendingWebRequests);


		if (storedData != null){
			System.println("storedData != null");
			App.getApp().setProperty("initialisation", false);
			App.getApp().setProperty("nbrFailedConnectionWithoutInit",0);
			App.getApp().setProperty("nbrFailedConnection",0);
			setProperty(type, storedData);
		} else {
			if (getProperty("initialisation")==false){
				System.println("getProperty(initialisation)==false");
				var nbrFailedConnection = App.getApp().getProperty("nbrFailedConnection");
				if (nbrFailedConnection == null){
					nbrFailedConnection = 1;
				} else{
					nbrFailedConnection = nbrFailedConnection+1;
				}
				App.getApp().setProperty("nbrFailedConnection",nbrFailedConnection);
			} else {
				System.println("getProperty(initialisation)==true");
				var nbrFailedConnectionWithoutInit = App.getApp().getProperty("nbrFailedConnectionWithoutInit");
				if (nbrFailedConnectionWithoutInit == null){
					nbrFailedConnectionWithoutInit = 1;
				} else{
					nbrFailedConnectionWithoutInit = nbrFailedConnectionWithoutInit+1;
				}
				App.getApp().setProperty("nbrFailedConnectionWithoutInit",nbrFailedConnectionWithoutInit);
			}
		}
		
		System.println("Ikigai-app-nbrFailedConnectionWithoutInit "+ App.getApp().getProperty("nbrFailedConnectionWithoutInit"));
	    System.println("Ikigai-app-nbrFailedConnection " + App.getApp().getProperty("nbrFailedConnection"));
	    
		Ui.requestUpdate();
	}

	// Return a formatted time dictionary that respects is24Hour and HideHoursLeadingZero settings.
	// - hour: 0-23.
	// - min:  0-59.
	function getFormattedTime(hour, min) {
		var amPm = "";

		if (!Sys.getDeviceSettings().is24Hour) {

			// #6 Ensure noon is shown as PM.
			var isPm = (hour >= 12);
			if (isPm) {
				
				// But ensure noon is shown as 12, not 00.
				if (hour > 12) {
					hour = hour - 12;
				}
				amPm = "p";
			} else {
				
				// #27 Ensure midnight is shown as 12, not 00.
				if (hour == 0) {
					hour = 12;
				}
				amPm = "a";
			}
		}

		// #10 If in 12-hour mode with Hide Hours Leading Zero set, hide leading zero. Otherwise, show leading zero.
		// #69 Setting now applies to both 12- and 24-hour modes.
		hour = hour.format(getProperty("HideHoursLeadingZero") ? INTEGER_FORMAT : "%02d");

		return {
			:hour => hour,
			:min => min.format("%02d"),
			:amPm => amPm
		};
	}
}
